package sned;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class LSH {
	private List<List<PCFGRule>> functions;
	private Map<String, List<Integer>> cache; 
	
	public LSH(List<PCFGTree> exemplars, int rule_count, int function_count){
		System.out.println("Constructing " + function_count + " LSH function from " + exemplars.size() + " an extracting " + rule_count + " rules.");
		
		List<PCFGRule> rules = new ArrayList<PCFGRule>();
		for(PCFGTree tree : exemplars){
			rules.addAll(tree.getAllRules()); 
		}
		
		//count rules
		Map<String, Integer> rule_counts = new HashMap<String, Integer>();
		for(PCFGRule rule : rules){
			if(rule.right.size()>1){
				rule_counts.put(rule.toString(), 0);
			}
		}
		for(PCFGRule rule : rules){
			if(rule.right.size()>1){
				String key = rule.toString(); 
				rule_counts.put(key, rule_counts.get(key)+1);
			}
		}
		
		//get border of count for reasonable rules
		List<Integer> counts = new ArrayList<Integer>();
		for(Integer count : rule_counts.values()){
			counts.add(count);
		}
		Collections.sort(counts);
		int border_count = (counts.size() < rule_count) ? 0 : counts.get(counts.size()-rule_count); 
	
		this.functions = new ArrayList<List<PCFGRule>>();
		for(int fc=0; fc < function_count ; fc++){
			List<PCFGRule> hf = new ArrayList<PCFGRule>();
			for(Map.Entry<String, Integer> entry : rule_counts.entrySet()){
				if(entry.getValue() >= border_count){
					//System.out.println(entry.getKey()+"="+entry.getValue());
					hf.add(new PCFGRule(entry.getKey())); 
				}
			}
			Collections.shuffle(hf); 
			this.functions.add(hf);
		}
		/*
		this.getHash(new Sentence("PP NP VP"));
		this.getHash(new Sentence("PP NP VP MD ADVP VP"));
		this.getHash(new Sentence("NNS NNS NNS NNS NNS NNS"));
		this.getHash(new Sentence("PP NNS NNS VP"));*/
	}
	
	public void buildCache(List<Sentence> sentences){
		System.out.println("Building LSH cache from " + sentences.size() + " sentences.");
		this.cache = new HashMap<String, List<Integer>>();
		
		for(int i=0; i<sentences.size() ; i++){
			String[] subhashes = this.getHash(sentences.get(i)).split(";");
			for(String subhash : subhashes){
				if(!this.cache.containsKey(subhash)){
					this.cache.put(subhash, new ArrayList<Integer>());
				}
				this.cache.get(subhash).add(i); 
			}
			if(i%500==0){
				System.out.println(i);
			}
		}
		System.out.println("LSH cache successfully built - contains " + this.cache.size() + " subhashes.");
	}
	
	//hash1;hash2;...;hashn where hashi = tag1,tag2,...tagni
	public String getHash(Sentence sen){
		String result="";
		int rule_count = this.functions.size(); 
		for(int rc=0; rc < rule_count ; rc++){
			if(rc>0){
				result += ";";
			}
			String ss = sen.toString();
			for(int i=0; i<this.functions.get(rc).size() ; i++){
				PCFGRule rule = this.functions.get(rc).get(i); 
				String rs = rule.getRightAsString();
				int pos = 0;
				while((pos = ss.indexOf(rs)) >= 0){
					//System.out.println(ss + " replace " + rs + " for " + rule.left);
					ss = ss.substring(0, pos) + rule.left + ss.substring(pos + rs.length());
				}
			}
			result += ss; 
		}
		//System.out.println("Hash of '"+sen+"' is '"+result+"'");
		return result;
	}
	
	public List<Integer> getCacheCollisions(Sentence sentence){
		String[] subhashes = this.getHash(sentence).split(";"); 
		List<Integer> result = new ArrayList<Integer>();
		for(String subhash : subhashes){
			if(this.cache.containsKey(subhash)){
				result.addAll(this.cache.get(subhash));
			}
		}
		return result; 
	}
	
	public static boolean hashEqual(String hash1, String hash2){
		List<String> t1 = Arrays.asList(hash1.split(";"));
		List<String> t2 = Arrays.asList(hash2.split(";"));
		for(String w1 : t1){
			for(String w2 : t2){
				if(w1.equals(w2)){
					return true;
				}
			}
		}
		return false;
	}
}
