
package topics;

import java.io.BufferedReader;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.TreeMap;
import utils.Counter;
import utils.Pair;

public class Topics {

    private static final String keysFile = "../wikidata/keys.txt";
    private static final String weightsFile = "../wikidata/wikidata_weights";
    private static final String countsFile = "../wikidata/wikidata_frequencies";

    int nextWordIndex = 0;
    int numTopics = 0;

    Map<String, Integer> wordToIndex = new TreeMap<String, Integer>();
    Map<Integer, String> indexToWord = new TreeMap<Integer, String>();

    Map<Integer, ArrayList<Double>> wordWeights = new TreeMap<Integer, ArrayList<Double>>();
    Map<Integer, Double> wordNormConsts = new TreeMap<Integer, Double>();
    ArrayList<Double> unknownWordWeights = new ArrayList<Double>();

    ArrayList<Double> topicWeights = new ArrayList<Double>();

    Counter<String> wordFreqs = new Counter<String>();

    public Topics() {
        this(keysFile, weightsFile, countsFile);
    }

    public Topics(String keysFile, String weightsFile, String countsFile) {
        loadKeys(keysFile);
        loadWeights(weightsFile);
        normalizeWordWeights();
        loadCounts(countsFile);
    }

    private void loadKeys(String keysFile) {
        System.out.println("topics > loading keys file...");
        double normConst = 0.0;
        try {
            BufferedReader in = new BufferedReader(new FileReader(keysFile));
            String line;
            while ((line = in.readLine()) != null) {
                String[] words = line.split("\\s");

                int topic = Integer.parseInt(words[0]);
                double weight = Double.parseDouble(words[1]);
                normConst += weight*weight;

                if (topic >= numTopics) numTopics = topic + 1;
                while (topicWeights.size() < numTopics) {
                    topicWeights.add(0.0);
                }

                topicWeights.set(topic, weight);
            }
            in.close();

            for (int i = 0; i < numTopics; ++i) {
                topicWeights.set(i, topicWeights.get(i) / normConst);
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        System.out.println("topics > loaded " + numTopics + " topics");
    }

    private void loadWeights(String weightsFile) {
        System.out.println("topics > loading weights file...");
        try {
            BufferedReader in = new BufferedReader(new FileReader(weightsFile));
            String line;
            while ((line = in.readLine()) != null) {
                String[] words = line.split("\\s");
                if (!wordToIndex.containsKey(words[1])) {
                    wordToIndex.put(words[1], nextWordIndex);
                    indexToWord.put(nextWordIndex, words[1]);
                    ++nextWordIndex;
                }

                int topic = Integer.parseInt(words[0]);
                int word = wordToIndex.get(words[1]);
                double weight = Double.parseDouble(words[2]);

                if (topic >= numTopics) numTopics = topic + 1;

                if (!wordWeights.containsKey(word)) {
                    wordWeights.put(word, new ArrayList<Double>());
                }
                while (wordWeights.get(word).size() < numTopics) {
                    wordWeights.get(word).add(0.0);
                }
                wordWeights.get(word).set(topic, weight);

                //if (!wordNormConsts.containsKey(word)) {
                //    wordNormConsts.put(word, 0.0);
                //}
                //wordNormConsts.put(word, wordNormConsts.get(word) + weight*weight);
            }
            in.close();

            for (int i = 0; i < numTopics; ++i) {
                unknownWordWeights.add(1.0 / (double) numTopics);
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        System.out.println("topics > loaded " + nextWordIndex + " words");
    }

    private void normalizeWordWeights() {
        System.out.println("topics > normalizing word weights...");
        for (Integer word : wordWeights.keySet()) {
            ArrayList<Double> weights = wordWeights.get(word);
            //double normConst = wordNormConsts.get(word);
            double normConst = 0.0;
            for (int i = 0; i < weights.size(); ++i) {
                normConst += weights.get(i)*weights.get(i);
            }
            for (int i = 0; i < weights.size(); ++i) {
                weights.set(i, weights.get(i) / Math.sqrt(normConst));
            }
        }
        /*for (Integer word : wordWeights.keySet()) {
            ArrayList<Double> weights = wordWeights.get(word);
            double normConst = 0.0;
            for (int i = 0; i < weights.size(); ++i) {
                normConst += weights.get(i);
            }
            System.out.println(word + " " + normConst);
        }*/
        System.out.println("topics > done normalizing");
    }

    private void loadCounts(String countsFile) {
        System.out.println("topics > loading frequencies file...");
        try {
            BufferedReader in = new BufferedReader(new FileReader(countsFile));
            String line;
            while ((line = in.readLine()) != null) {
                String[] words = line.split("\\s");

                String word = words[0];
                double freq = Double.parseDouble(words[1]);

                wordFreqs.set(word, freq);
            }
            in.close();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        System.out.println("topics > loaded frequencies");
    }

    private ArrayList<Double> getWeightsVector(String word) {
        if (wordToIndex.containsKey(word)) {
            int index = wordToIndex.get(word);
            if (wordWeights.containsKey(index)) {
                return wordWeights.get(index);
            }
        }
        return unknownWordWeights;
    }

    public double getFreqNormConst(String word1, String word2) {
        double c1 = wordFreqs.get(word1);
        double c2 = wordFreqs.get(word2);
        return 1.0 / (Math.sqrt(1.0 + Math.abs(c2 - c1) / Math.min(c1, c2)));
    }

    public double getDotProduct(String word1, String word2) {
        ArrayList<Double> v1 = getWeightsVector(word1);
        ArrayList<Double> v2 = getWeightsVector(word2);
        int size = Math.min(v1.size(), v2.size());
        double dotProd = 0.0;
        for (int i = 0; i < size; ++i) {
            dotProd += v1.get(i) * v2.get(i);// * topicWeights.get(i);
        }
        return dotProd * getFreqNormConst(word1, word2);
    }

    public void printDotProduct(String word1, String word2) {
        System.out.println("topics > dot product for " + word1 + " and " + word2);
        System.out.println(getDotProduct(word1, word2));
    }

    public void listNearest(String word1, int count) {
        System.out.println("topics > searching nearest words for " + word1 + "...");
        PriorityQueue<Pair<Double, String>> nearest = new PriorityQueue<Pair<Double, String>>();
        for (String word2 : wordToIndex.keySet()) {
            nearest.add(new Pair<Double, String>(1.0 - getDotProduct(word1, word2), word2));
        }
        for (int i = 0; i < count; ++i) {
            Pair<Double, String> pair = nearest.poll();
            System.out.println(pair.second + "\t" + (1.0 - pair.first));
        }
    }

}
